/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.kafka.connect.runtime.rest.prometheus;

import io.prometheus.client.Gauge;
import io.prometheus.client.exporter.MetricsServlet;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.management.MBeanServer;
import javax.management.ObjectInstance;
import javax.management.ObjectName;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.lang.management.ManagementFactory;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class PrometheusConnectMetricsServlet extends MetricsServlet {

    private static final Logger log = LoggerFactory.getLogger(PrometheusConnectMetricsServlet.class);
    private static final String CONNECT_PREFIX = "kafka_connect_";
    private static final String CONNECT_WORKER_PREFIX = "kafka_connect_worker_";
    private static final String CONNECT_WORKER_REBALANCE_PRFIX = "kafka_connect_worker_rebalance_";

    // Kafka Connect JMX Types
    private static final String WORKER_METRICS = "kafka.connect:type=connect-worker-metrics";
    private static final String WORKER_REBALANCE_METRICS = "kafka.connect:type=connect-worker-rebalance-metrics";
    private static final String CONNECTOR_TASK_METRICS = "kafka.connect:type=connector-task-metrics";
    private static final String SOURCE_TASK_METRICS = "kafka.connect:type=source-task-metrics";
    private static final String SINK_TASK_METRICS = "kafka.connect:type=sink-task-metrics";
    private static final String TASK_ERROR_METRICS = "kafka.connect:type=task-error-metrics";

    private static final Pattern TASK_METRICS_PATTERN = Pattern.compile("^kafka.connect:type=(.+)-metrics,connector=(.+),task=(.+)$");

    // Kafka Connect JMX Attributes
    private static final Set<String> WORKER_METRICS_ATTRIBUTES = new HashSet<>(Arrays.asList("connector-count", "connector-startup-attempts-total",
        "connector-startup-failure-total", "task-count", "task-startup-attempts-total", "task-startup-failure-total", "task-startup-success-total"));
    private static final Set<String> WORKER_REBALANCE_METRICS_ATTRIBUTES = new HashSet<>(Arrays.asList("completed-rebalances-total",
        "rebalance-avg-time-ms", "rebalance-max-time-ms", "epoch", "rebalancing", "time-since-last-rebalance-ms"));
    private static final Set<String> CONNECTOR_TASK_METRICS_ATTRIBUTES = new HashSet<>(Arrays.asList("offset-commit-avg-time-ms",
        "offset-commit-max-time-ms", "offset-commit-success-percentage", "offset-commit-failure-percentage", "running-ratio",
        "pause-ratio", "batch-size-avg", "batch-size-max"));
    private static final Set<String> SOURCE_TASK_METRICS_ATTRIBUTES = new HashSet<>(Arrays.asList("source-record-write-rate",
        "poll-batch-avg-time-ms", "poll-batch-max-time-ms", "source-record-write-total", "source-record-poll-total",
        "source-record-poll-rate", "source-record-active-count", "source-record-active-count-max", "source-record-active-count-avg"));
    private static final Set<String> SINK_TASK_METRICS_ATTRIBUTES = new HashSet<>(Arrays.asList("partition-count",
        "sink-record-lag-max", "sink-record-send-rate", "put-batch-avg-time-ms"));
    private static final Set<String> TASK_ERROR_METRICS_ATTRIBUTES = new HashSet<>(Arrays.asList("last-error-timestamp",
        "total-record-failures", "total-record-errors", "total-records-skipped", "total-retries", "total-errors-logged"));

    private final Map<String, Gauge> allGauges = new HashMap<>();

    @Override
    protected void doGet(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {
        createPrometheusMetrics();
        super.doGet(request, response);
    }

    @Override
    protected void doOptions(HttpServletRequest req, HttpServletResponse resp) {
        resp.setHeader("Allow", "GET, HEAD, POST, OPTIONS");
    }

    void createPrometheusMetrics() {
        for (Gauge gauge : allGauges.values()) {
            gauge.clear();
        }

        MBeanServer server = ManagementFactory.getPlatformMBeanServer();
        Set<ObjectInstance> instances = server.queryMBeans(null, null);

        for (ObjectInstance instance : instances) {
            handleMetrics(server, instance.getObjectName());
        }
    }

    Map<String, Gauge> getAllGauges() {
        return allGauges;
    }

    private void handleMetrics(MBeanServer server, ObjectName objectName) {
        String objectNameString = objectName.toString();
        if (objectNameString.equals(WORKER_METRICS)) {
            handleWorkerMetrics(server, objectName, CONNECT_WORKER_PREFIX, WORKER_METRICS_ATTRIBUTES);
        } else if (objectNameString.equals(WORKER_REBALANCE_METRICS)) {
            handleWorkerMetrics(server, objectName, CONNECT_WORKER_REBALANCE_PRFIX, WORKER_REBALANCE_METRICS_ATTRIBUTES);
        } else if (objectNameString.startsWith(CONNECTOR_TASK_METRICS)) {
            handleTaskMetrics(server, objectName, CONNECTOR_TASK_METRICS_ATTRIBUTES);
        } else if (objectNameString.startsWith(SOURCE_TASK_METRICS)) {
            handleTaskMetrics(server, objectName, SOURCE_TASK_METRICS_ATTRIBUTES);
        } else if (objectNameString.startsWith(SINK_TASK_METRICS)) {
            handleTaskMetrics(server, objectName, SINK_TASK_METRICS_ATTRIBUTES);
        } else if (objectNameString.startsWith(TASK_ERROR_METRICS)) {
            handleTaskMetrics(server, objectName, TASK_ERROR_METRICS_ATTRIBUTES);
        }
    }

    private void handleWorkerMetrics(MBeanServer server, ObjectName objectName, String prefix, Set<String> attributes) {
        for (String attributeName : attributes) {
            double value;

            try {
                value = Double.parseDouble(server.getAttribute(objectName, attributeName).toString());
            } catch (Exception e) {
                log.debug("Exception retrieving {} from {}", attributeName, objectName, e);
                continue;
            }

            String builder = prefix + attributeName;
            Gauge gauge = getGauge(builder.replace("-", "_"));
            gauge.set(value);
        }
    }

    private void handleTaskMetrics(MBeanServer server, ObjectName objectName, Set<String> attributes) {
        Matcher matcher = TASK_METRICS_PATTERN.matcher(objectName.toString());

        if (!matcher.matches()) {
            return;
        }

        for (String attributeName : attributes) {
            double value;

            try {
                value = Double.parseDouble(server.getAttribute(objectName, attributeName).toString());
            } catch (Exception e) {
                log.debug("Exception retrieving {} from {}", attributeName, objectName, e);
                continue;
            }

            String type = matcher.group(1);
            String connector = matcher.group(2);
            String task = matcher.group(3);

            String builder = CONNECT_PREFIX + type + "_" + attributeName;
            Gauge gauge = getGauge(builder.replace("-", "_"), "connector", "task");
            gauge.labels(connector, task).set(value);
        }
    }

    private Gauge getGauge(String name, String... labels) {
        Gauge gauge = allGauges.get(name);

        if (gauge == null) {
            Gauge.Builder builder = Gauge.build().name(name);

            if (labels.length > 0) {
                builder.labelNames(labels);
            }

            gauge = builder.help("Kafka Connect Metrics").register();
            allGauges.put(name, gauge);
        }

        return gauge;
    }

}
